#!/usr/bin/env python3
#
# Copyright 2020 Graviti. All Rights Reserved.
#

"""This file contains checkhealth related functions."""

from typing import Union

from .dataset import Dataset, FusionDataset


def checkhealth(dataset: Union[Dataset, FusionDataset]) -> bool:
    pass
